import asyncio
import contextvars
import os
import sys
from concurrent.futures import ThreadPoolExecutor
from functools import partial, wraps
from time import time
from typing import Callable, Coroutine, Iterable, ParamSpec, TypeVar
from .iter_utils import as_list
from .processing_utils import noop

__all__ = [
  "as_async", "async_limit",
  "clear_console", "console_link",
  "default_executor", "default_workers",
  "raise_if_interrupt", "roll_tasks",
  "measure_duration",
]

P = ParamSpec("P")
R = TypeVar("R")
Coro = Coroutine[object, object, R]

default_workers = min(32, (os.cpu_count() or 1) + 4)
default_executor = ThreadPoolExecutor(max_workers=default_workers)

def raise_if_interrupt():
  if sys.exc_info()[0] in (KeyboardInterrupt, SystemExit):
    raise

def clear_console() -> None:
  os.system("cls" if os.name == "nt" else "clear")

def console_link(text: str, url: str) -> str:
  return f"\033]8;;{url}\033\\{text}\033]8;;\033\\"

async def worker[T](task: Coro[T], semaphore: asyncio.Semaphore, update=noop) -> T:
  async with semaphore:
    result = await task
    update()
    return result

async def roll_tasks[T](tasks: Iterable[Coro[T]], workers=default_workers, progress=False) -> list[T]:
  semaphore = asyncio.Semaphore(workers)
  if not progress:
    return await asyncio.gather(*[worker(task, semaphore) for task in tasks])

  from tqdm import tqdm
  tasks = as_list(tasks)
  with tqdm(total=len(tasks)) as pbar:
    update = partial(pbar.update, 1)
    return await asyncio.gather(*[worker(task, semaphore, update) for task in tasks])

def as_async(workers: int | ThreadPoolExecutor = default_executor) -> Callable[[Callable[P, R]], Callable[P, Coro[R]]]:
  executor = ThreadPoolExecutor(max_workers=workers) if isinstance(workers, int) else workers

  def on_fn(func: Callable[P, R]) -> Callable[P, Coro[R]]:
    @wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
      loop = asyncio.get_running_loop()
      ctx = contextvars.copy_context()
      fn_call = partial(ctx.run, func, *args, **kwargs)
      return await loop.run_in_executor(executor, fn_call)
    return wrapper
  return on_fn

def async_limit(workers=default_workers) -> Callable[[Callable[P, Coro[R]]], Callable[P, Coro[R]]]:
  semaphore = asyncio.Semaphore(workers)

  def on_fn(func: Callable[P, Coro[R]]) -> Callable[P, Coro[R]]:
    @wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
      async with semaphore:
        return await func(*args, **kwargs)
    return wrapper
  return on_fn

active_mds = []

class measure_duration:
  def __init__(self, name):
    self.name = name
    active_mds.append(self)

  def __enter__(self):
    self.start = time()

  def __exit__(self, *_):
    duration = round(time() - self.start, 4)
    depth = len(active_mds) - 1
    indent = "──" * depth + " " * (depth > 0)
    text = f"{self.name}: {duration} seconds"
    print(indent + text)
    active_mds.remove(self)



from contextlib import contextmanager
from itertools import chain, islice
from typing import Any, Iterable, Literal, Self, Sequence, overload
from .dict_utils import dict_firsts

__all__ = [
  "as_list",
  "at",
  "chunked",
  "distinct_by", "distinct", "drop_none",
  "first", "flatten",
  "interleave", "intersect",
  "list_split",
  "move_value",
  "partition",
  "reversed_enumerate",
  "seekable", "sort_by",
  "transpose",
]

def as_list[T](iterable: Iterable[T]) -> list[T]:
  return iterable if isinstance(iterable, list) else list(iterable)

def at[T, U](values: Sequence[T], index: int, default: U = None) -> T | U:
  try:
    return values[index]
  except IndexError:
    return default

def first[T](iterable: Iterable[T]) -> T | None:
  return next(iter(iterable), None)

def drop_none[T](iterable: Iterable[T | None]) -> list[T]:
  return [x for x in iterable if x is not None]

def distinct[T](iterable: Iterable[T]) -> list[T]:
  return list(dict.fromkeys(iterable))

def distinct_by[T](pairs: Iterable[tuple[object, T]]) -> list[T]:
  return list(dict_firsts(pairs).values())

def sort_by[T](pairs: Iterable[tuple[Any, T]], reverse=False) -> list[T]:
  pairs = sorted(pairs, key=lambda p: p[0], reverse=reverse)
  return [v for _, v in pairs]

def move_value[T](iterable: Iterable[T], from_i: int, to_i: int) -> list[T]:
  values = list(iterable)
  values.insert(to_i, values.pop(from_i))
  return values

def reversed_enumerate[T](values: Sequence[T] | tuple[T, ...]) -> Iterable[tuple[int, T]]:
  return zip(range(len(values))[::-1], reversed(values))

def intersect[T](*iterables: Iterable[T]) -> list[T]:
  return list(set.intersection(*map(set, iterables)))

def interleave[T](*iterables: Iterable[T]) -> list[T]:
  return flatten(zip(*iterables))

def list_split[T](iterable: Iterable[T], sep: T) -> list[list[T]]:
  values = [sep, *iterable, sep]
  split_at = [i for i, x in enumerate(values) if x is sep]
  ranges = list(zip(split_at[0:-1], split_at[1:]))
  return [values[start + 1:end] for start, end in ranges]

def partition[T](iterable: Iterable[tuple[bool, T]]) -> tuple[list[T], list[T]]:
  true_values, false_values = [], []
  for predicate, value in iterable:
    if predicate:
      true_values.append(value)
    else:
      false_values.append(value)
  return true_values, false_values

class seekable[T]:
  def __init__(self, iterable: Iterable[T]):
    self.index = 0
    self.source = iter(iterable)
    self.sink: list[T] = []

  def __iter__(self):
    return self

  def __next__(self) -> T:
    if len(self.sink) > self.index:
      item = self.sink[self.index]
    else:
      item = next(self.source)
      self.sink.append(item)
    self.index += 1
    return item

  def __bool__(self):
    return bool(self.lookahead(1))

  def clear(self):
    self.sink[:self.index] = []
    self.index = 0

  def seek(self, index: int) -> Self:
    remainder = index - len(self.sink)
    if remainder > 0:
      next(islice(self, remainder, remainder), None)
    self.index = max(0, min(index, len(self.sink)))
    return self

  def step(self, count: int) -> Self:
    return self.seek(self.index + count)

  @contextmanager
  def freeze(self):
    def commit(offset: int = 0):
      nonlocal initial_index
      initial_index = self.index + offset
    initial_index = self.index
    try:
      yield commit
    finally:
      self.seek(initial_index)

  def lookahead(self, count: int) -> list[T]:
    with self.freeze():
      return list(islice(self, count))

@overload
def chunked[T](values: Iterable[T], *, num_chunks: int, chunk_size=None) -> list[list[T]]: ...
@overload
def chunked[T](values: Iterable[T], *, num_chunks=None, chunk_size: int) -> list[list[T]]: ...
def chunked(values, *, num_chunks=None, chunk_size=None):
  values = as_list(values)
  if isinstance(num_chunks, int):
    chunk_size = (len(values) / num_chunks).__ceil__()
  elif isinstance(chunk_size, int):
    num_chunks = (len(values) / chunk_size).__ceil__()
  assert isinstance(num_chunks, int) and isinstance(chunk_size, int)
  return [values[i * chunk_size:(i + 1) * chunk_size] for i in range(num_chunks)]

@overload
def flatten[T](iterable: Iterable[T], depth: Literal[0]) -> list[T]: ...
@overload
def flatten[T](iterable: Iterable[Iterable[T]], depth: Literal[1] = 1) -> list[T]: ...
@overload
def flatten[T](iterable: Iterable[Iterable[Iterable[T]]], depth: Literal[2]) -> list[T]: ...
@overload
def flatten[T](iterable: Iterable[Iterable[Iterable[Iterable[T]]]], depth: Literal[3]) -> list[T]: ...
@overload
def flatten[T](iterable: Iterable[Iterable[Iterable[Iterable[Iterable[T]]]]], depth: Literal[4]) -> list[T]: ...
@overload
def flatten(iterable: Iterable, depth: int) -> list: ...
def flatten(iterable: Iterable, depth: int = 1) -> list:
  for _ in range(depth):
    iterable = chain.from_iterable(iterable)
  return list(iterable)

@overload
def transpose[T1, T2](tuples: Iterable[tuple[T1, T2]], default_num_returns=0) -> tuple[list[T1], list[T2]]: ...
@overload
def transpose[T1, T2, T3](tuples: Iterable[tuple[T1, T2, T3]], default_num_returns=0) -> tuple[list[T1], list[T2], list[T3]]: ...
@overload
def transpose[T1, T2, T3, T4](tuples: Iterable[tuple[T1, T2, T3, T4]], default_num_returns=0) -> tuple[list[T1], list[T2], list[T3], list[T4]]: ...
@overload
def transpose[T1, T2, T3, T4, T5](tuples: Iterable[tuple[T1, T2, T3, T4, T5]], default_num_returns=0) -> tuple[list[T1], list[T2], list[T3], list[T4], list[T5]]: ...
@overload
def transpose[T](tuples: Iterable[tuple[T, ...]], default_num_returns=0) -> tuple[list[T], ...]: ...
def transpose(tuples: Iterable[tuple], default_num_returns=0) -> tuple[list, ...]:
  output = tuple(zip(*tuples))
  if not output:
    return ([],) * default_num_returns
  return tuple(map(list, output))



import re
from typing import Any, Callable, Iterable, overload

__all__ = [
  "clamp",
  "df_from_array",
  "for_each",
  "noop",
  "str_filterer", "StrFilter",
]

def noop() -> None:
  pass

def for_each[T](func: Callable[[T], Any], iterable: Iterable[T]) -> None:
  for item in iterable:
    func(item)

@overload
def clamp(value: int, low: int, high: int) -> int: ...
@overload
def clamp(value: float, low: float, high: float) -> float: ...
def clamp(value: float, low: float, high: float) -> float:
    return max(low, min(value, high))

def _cat_tile(cats, n_tile):
  import numpy as np
  return cats[np.tile(np.arange(len(cats)), n_tile)]

def df_from_array(
  value_cols: dict[str, Any],
  dim_labels: list[tuple[str, list[str | int | float]]],
  indexed=False,
):
  import numpy as np
  import pandas as pd
  dim_sizes = np.array([len(labels) for _, labels in dim_labels])
  assert all(array.shape == tuple(dim_sizes) for array in value_cols.values())
  array_offsets = [
    (dim_sizes[i + 1:].prod(), dim_sizes[:i].prod())
    for i in range(len(dim_sizes))
  ]
  category_cols = {
    dim: _cat_tile(pd.Categorical(labels).repeat(repeats), tiles)
    for (dim, labels), (repeats, tiles) in zip(dim_labels, array_offsets)
  }
  value_cols = {name: array.reshape(-1) for name, array in value_cols.items()}
  df = pd.DataFrame({**category_cols, **value_cols}, copy=False)
  if indexed:
    df = df.set_index([name for name, _ in dim_labels])
  return df

StrFilter = Callable[[str], bool]

def str_filterer(
  include_patterns: list[re.Pattern[str]] = [],
  exclude_patterns: list[re.Pattern[str]] = [],
) -> StrFilter:
  def str_filter(string: str) -> bool:
    if any(pattern.search(string) for pattern in exclude_patterns):
      return False
    if not include_patterns:
      return True
    return any(pattern.search(string) for pattern in include_patterns)

  return str_filter



from typing import Any, Callable, Iterable, overload
from .type_utils import as_any

__all__ = [
  "deep_dict_pairs", "deepen_dict", "dict_by", "dict_firsts",
  "flatten_dict",
  "get_at", "group",
  "key_of",
  "map_dict", "merge_dicts",
  "omit",
  "pick",
  "tuple_by",
]

def merge_dicts[T, K](*dicts: dict[K, T]) -> dict[K, T]:
  if len(dicts) == 1:
    return dicts[0]
  result = {}
  for d in dicts:
    result |= d
  return result

def omit[T, K](d: dict[K, T], keys: Iterable[K]) -> dict[K, T]:
  if keys:
    d = dict(d)
    for key in keys:
      del d[key]
  return d

def pick[T, K](d: dict[K, T], keys: Iterable[K]) -> dict[K, T]:
  return {key: d[key] for key in keys}

def dict_by[T, K](keys: Iterable[K], values: Iterable[T]) -> dict[K, T]:
  return dict(zip(keys, values))

def tuple_by[T, K](d: dict[K, T], keys: Iterable[K]) -> tuple[T, ...]:
  return tuple(d[key] for key in keys)

def map_dict[T, U, K](fn: Callable[[T], U], d: dict[K, T]) -> dict[K, U]:
  return {key: fn(value) for key, value in d.items()}

def key_of[T, U](dicts: Iterable[dict[T, U]], key: T) -> list[U]:
  return [d[key] for d in dicts]

def get_at[T](d: dict, keys: Iterable[Any], default: T) -> T:
  try:
    for key in keys:
      d = d[key]
  except KeyError:
    return default
  return as_any(d)

def dict_firsts[T, K](pairs: Iterable[tuple[K, T]]) -> dict[K, T]:
  result: dict[K, T] = {}
  for key, value in pairs:
    result.setdefault(key, value)
  return result

def group[T, K](pairs: Iterable[tuple[K, T]]) -> dict[K, list[T]]:
  values_by_key = {}
  for key, value in pairs:
    values_by_key.setdefault(key, []).append(value)
  return values_by_key

def deep_dict_pairs(d, prefix=()):
  for key, value in d.items():
    if not isinstance(value, dict) or value == {}:
      yield prefix + (key,), value
    else:
      yield from deep_dict_pairs(value, prefix + (key,))

def flatten_dict(deep_dict: dict, prefix=()) -> dict:
  return dict(deep_dict_pairs(deep_dict, prefix))

@overload
def deepen_dict[K1, U](d: dict[tuple[K1], U]) -> dict[K1, U]: ...
@overload
def deepen_dict[K1, K2, U](d: dict[tuple[K1, K2], U]) -> dict[K1, dict[K2, U]]: ...
@overload
def deepen_dict[K1, K2, K3, U](d: dict[tuple[K1, K2, K3], U]) -> dict[K1, dict[K2, dict[K3, U]]]: ...
@overload
def deepen_dict[K1, K2, K3, K4, U](d: dict[tuple[K1, K2, K3, K4], U]) -> dict[K1, dict[K2, dict[K3, dict[K4, U]]]]: ...
@overload
def deepen_dict[K1, K2, K3, K4, K5, U](d: dict[tuple[K1, K2, K3, K4, K5], U]) -> dict[K1, dict[K2, dict[K3, dict[K4, dict[K5, U]]]]]: ...
@overload
def deepen_dict[K1, K2, K3, K4, K5, K6, U](d: dict[tuple[K1, K2, K3, K4, K5, K6], U]) -> dict[K1, dict[K2, dict[K3, dict[K4, dict[K5, dict[K6, U]]]]]]: ...
def deepen_dict(d: dict[tuple[Any, ...], Any]) -> dict:
  output = {}
  if () in d:
    return d[()]
  for (*tail, head), value in d.items():
    curr = output
    for key in tail:
      curr = curr.setdefault(key, {})
    curr[head] = value
  return output



import json
from pathlib import Path
from typing import Any, Iterable

__all__ = [
  "clear_directory",
  "empty_dirs",
  "read_json",
  "write_json",
]

default_sentinel = object()

def read_json(path: Path, default=default_sentinel) -> Any:
  if default is not default_sentinel and not path.exists():
    return default
  with path.open("r") as f:
    return json.load(f)

def write_json(path: Path, obj: object, indent: None | int = None) -> None:
  with path.open("w") as f:
    separators = (",", ":") if indent is None else None
    return json.dump(obj, f, indent=indent, separators=separators)

def empty_dirs(path: Path) -> Iterable[Path]:
  nonempty_count = 0
  for child in path.iterdir():
    nonempty_count += 1
    if child.is_dir():
      for grand_child in empty_dirs(child):
        yield grand_child
        nonempty_count -= child == grand_child
  if nonempty_count == 0:
    yield path

def clear_directory(path: Path):
  if path.is_dir():
    for file in path.glob("**/.DS_Store"):
      file.unlink()
    for directory in empty_dirs(path):
      directory.rmdir()



from typing import Any

__all__ = [
  "as_any",
  "ensure_tuple",
  "non_none",
]

def as_any(obj: Any) -> Any:
  return obj

def non_none[T](obj: T | None) -> T:
  assert obj is not None
  return obj

def ensure_tuple[T](value: T | tuple[T, ...]) -> tuple[T, ...]:
  return value if isinstance(value, tuple) else (value,)
